
--  Untyped lambda terms with named variables (nats).

-- # import prelude
-- # import nat
-- # import list
-- # import nateq
-- # import id
-- # import bijection
-- # import prop

-- The terms --------------------------------------------------------------

infixl 20 @

Name = Nat

data Term : Set where
  var : Name -> Term
  (@) : Term -> Term -> Term
  λ   : Name -> Term -> Term

lam = λ

-- α-equality -------------------------------------------------------------

NameMap = PartialBijection Name Name

data AlphaEqual : NameMap -> Term -> Term -> Type where

  varEq (Φ : NameMap)(x,y : Name) :

	      pair x y ∈ Φ
    -> ----------------------------
       AlphaEqual Φ (var x) (var y)


  appEq (Φ : NameMap)(s1,s2,t1,t2 : Term) :

       AlphaEqual Φ s1 s2  ->  AlphaEqual Φ t1 t2
    -> ------------------------------------------
	    AlphaEqual Φ (s1 @ t1) (s2 @ t2)


  λEq (Φ, Φ' : NameMap)(x, y : Name)(t1, t2 : Term) :
    
       Extends Φ' Φ (pair x y)  ->  AlphaEqual Φ' t1 t2
    -> ------------------------------------------------
	      AlphaEqual Φ (λ x t1) (λ y t2)

-- Two terms are (top-level) alpha equal if they are alpha equal under the
-- identity mapping.
(=~=) : Term -> Term -> Type
s =~= t = AlphaEqual identity s t

αRefl (t:Term) : t =~= t
αRefl {var x} = varEq {identity} refl
αRefl {s @ t} = appEq αRefl αRefl
αRefl {λ x t} = λEq {identity} {identity} idExtends αRefl

αSym' (Φ : NameMap)(s,t:Term) : AlphaEqual Φ s t -> AlphaEqual (inverse Φ) t s
αSym' (varEq el)	= varEq $ elemInverse el
αSym' (appEq eqs eqt)	= appEq (αSym' eqs) (αSym' eqt)
αSym' (λEq ext eqt)	= λEq (extInverse ext) (αSym' eqt)

-- We can't say αSym = αSym' since inverse identity != identity
αSym (s, t : Term) : s =~= t -> t =~= s
αSym (varEq el)		    = varEq {identity} (sym el)
αSym (appEq eqs eqt)	    = appEq (αSym eqs) (αSym eqt)
αSym (λEq {_} {Φ'} ext eqt) = λEq {identity} {inverse Φ'}
				  (extIdInverse ext)
				  (αSym' eqt)

-- Free variables ---------------------------------------------------------

data Free (x:Name) : Term -> Set where
  varFree			:			 Free x (var x)
  appFree1 (s, t : Term)	:	     Free x s -> Free x (s @ t)
  appFree2 (s, t : Term)	:	     Free x t -> Free x (s @ t)
  λFree    (y : Name)(t : Term) : y =/= x -> Free x t -> Free x (λ y t)

data NotFree (x:Name) : Term -> Set where
  varNotFree   (y : Name)	    : y =/= x ->		    NotFree x (var y)
  appNotFree   (s, t : Term)	    : NotFree x s -> NotFree x t -> NotFree x (s @ t)
  boundNotFree (t : Term)	    :				    NotFree x (λ x t)
  λNotFree     (y : Name)(t : Term) : NotFree x t ->		    NotFree x (λ y t)

freeOrNot : (x:Name) -> (t:Term) -> Free x t \/ NotFree x t
freeOrNot x (var y) with natCmp y x =>
  x (var x) sameNat	  = orI1 varFree
  x (var y) (diffNats d)  = orI2 (varNotFree d)
freeOrNot x (s @ t) with freeOrNot x s =>
  x (s @ t) (orI1 f)  = orI1 (appFree1 f)
  x (s @ t) (orI2 nf) with freeOrNot x t =>
    x (s @ t) (orI2 _)	 (orI1 f)   = orI1 (appFree2 f)
    x (s @ t) (orI2 nfs) (orI2 nft) = orI2 (appNotFree nfs nft)
freeOrNot x (λ y t) with natCmp y x =>
  x (λ x t) sameNat	  = orI2 boundNotFree
  x (λ y t) (diffNats _) with freeOrNot x t =>
    x (λ y t) (diffNats d) (orI1 f)  = orI1 (λFree d f)
    x (λ y t) (diffNats d) (orI2 nf) = orI2 (λNotFree nf)

-- Substitution -----------------------------------------------------------

data Subst (x : Name)(s : Term) : Term -> Term -> Set where
  substThisVar			 :	      Subst x s (var x) s
  substOtherVar	(y : Name)	 : x =/= y -> Subst x s (var y) (var y)
  substλ (y : Name)(t,t' : Term) : x =/= y -> NotFree y s -> Subst x s t t' ->
					      Subst x s (λ y t) (λ y t')
  substApp (t,t',u,u' : Term)	 : Subst x s t t' -> Subst x s u u' ->
					      Subst x s (t @ u) (t' @ u')

-- Reduction --------------------------------------------------------------

data (-->) : Term -> Term -> Set where
  β (x : Name)(s, t, u : Term) : Subst x t s u -> λ x s @ t --> u
  rapp1 (s,s',t : Term)	       : s --> s' -> s @ t --> s' @ t
  rapp2 (s,t,t' : Term)	       : t --> t' -> s @ t --> s @ t'
  rλ (x : Name)(t,t' : Term)   : t --> t' -> λ x t --> λ x t'

-- Modulo α-equivalence
data (~~>) (s, t : Term) : Type where

  rα : (s' : Term)

	-> s =~= s'  ->  s' --> t
	-> ----------------------
		  s ~~> t

-- Confluence -------------------------------------------------------------

localConfluence :
    (u, u1, u2 : Term) -> u ~~> u1 -> u ~~> u2 ->
    { v : Term; u1 ~~> v; u2 ~~> v }

-- rapp1, rapp2
-- localConfluence (u @ u') (u1 @ u1') (u2 @ u2')
-- 		(rα (s @ z) (s1 @ z) (αEq Φ1 idΦ1 (appEq eq_us eq_u'z))
-- 				     (αEq Φ2 idΦ2 (appEq eq_u1s1 eq_u1'z))
-- 				     (rapp1 r1)
-- 		)
-- 		(rα (w @ t) (w @ t1) (αEq Φ3 idΦ3 (appEq eq_uw eq_u't))
-- 				     (αEq Φ4 idΦ4 (appEq eq_u2w eq_u2't1))
-- 				     (rapp2 r2)
-- 		)
--   = (s1 @ t1, rα (u1 @ t) (s1 @ t1) assert-plugin() assert-plugin() assert-plugin(), assert-plugin())


