{-| Plugin for debug output. Prints the value and type of terms.
-}
module Plugins.Debug where

import Prelude hiding	    ( print, putStr, putStrLn )
import IO.Output

import Control.Monad.Trans
import Text.PrettyPrint

import Syntax.Internal
import Syntax.Pretty
import Syntax.Values

import TypeChecker
import TypeChecker.Monad
import TypeChecker.UnfoldConstants
import TypeChecker.Values
import TypeChecker.Values.DeBruijn
import TypeChecker.Monad.Errors
import TypeChecker.Errors

import Utils.Error
import Utils.Monads

import Plugins

plugin :: SomePlugin
plugin = mkPlugin
    Plugin  { plName		= "debug"
            , plHelp            = debugHelp
	    , plInit		= \_ -> return []
	    , plFinal		= debugFinal
	    , plCheck		= debugCheck
	    , plInfer		= debugInfer
	    , plOptions		= simpleOptions []
	    , plDefaultOptions	= ()
	    }

debugFinal _ [] = return ()
debugFinal _ vs =
    do	liftIO $ putStrLn $ "=== Debug output " ++ replicate 53 '='
	ds <- mapM pr $ reverse vs
	liftIO $ print $ nest 2 $ vcat ds
    where
	pr (v,t) =
	    do	v' <- unfoldAllConstants v
		t' <- unfoldAllConstants $ unEl t
		return $
		    sep [ pretty v
			, nest 2 $ text "value :" <+> pretty v'
			, nest 2 $ text "type  :" <+> pretty t
			, nest 2 $ text "      :" <+> pretty t'
			]

debugCheck _ vs _ [e] t =
    do	v <- checkType e t
	v' <- unBruijn v
	t' <- typeMapM unBruijn t
	return (v, (v',t'):vs)
debugCheck _ _ _ es _ = failure $ customError $ "debug-plugin expects 1 argument, not " ++ show (length es)

debugInfer _ vs _ [e] =
    do	(v,t) <- inferType e
	v' <- unBruijn v
	t' <- typeMapM unBruijn t
	return (v, t, (v',t'):vs)
debugInfer _ _ _ es = failure $ customError $ "debug-plugin expects 1 argument, not " ++ show (length es)

debugHelp = show $
    vcat    [ text "Provides facilities for printing the value and type of terms."
	    , nest 2 $ fsep $ pwords $
		    "debug-plugin(e) behaves as e but after typechecking the " ++
		    "value and type of e is printed."
	    , text ""
	    ]

