
module TypeChecker.Context where

import Utils.Map

import Syntax.Internal
import Syntax.Values

-- | A context is a map from variables to either types or values.
--   The type checking monad 'Monad.TC' keeps track of the current context
--   as part of the 'Monad.Environment'.
--   Note that defined things doesn't count to the deBuijn indices. For
--   instance: If @Gamma = [("x",Abstract t1),("y",Defined v),("z",Abstract t2)]@
--   @VBId 0@ refers to @z@ and @VBId 1@ refers to @x@. No value can refer to @y@.
type Context = [(Name, ContextEntry)]

data ContextEntry
	= Abstract  Type
	| Defined   Value Type

findInContext :: Name -> Context -> Maybe (Int, ContextEntry)
findInContext x e = find 0 x e
    where
	find _ _ [] = Nothing
	find n x ((y, Abstract t):c)
	    | x == y	= return (n, Abstract t)
	    | otherwise	= find (n + 1) x c
	find n x ((y, Defined v t):c)
	    | x == y	= return (n, Defined v t)
	    | otherwise	= find n x c

nofBoundVars :: Context -> Int
nofBoundVars gamma = length [ () | (_, Abstract _) <- gamma ]

-- upContext :: Context -> Context
-- upContext = fmap up'
--     where
-- 	up' (Abstract t) = Abstract $ up t
-- 	up' (Defined v)	 = Defined  $ up v

