
module Utils.Lists where

import Data.List
import Data.Maybe

-- | A variant of 'groupBy' that divides a list into constant chunks w.r.t.
--   a given function. For instance:
--
--   > chunks id "aaabccdddeaa" = ["aaa","b","cc","ddd","e","aa"]
chunks :: Eq b => (a -> b) -> [a] -> [[a]]
chunks f xs =
    map (map snd)
	$ groupBy (\ (v,_) (w,_) -> v == w)
	$ map (\x -> (f x, x)) xs

-- | Same as 'chunks' but sorts the list first.
classify :: Ord b => (a -> b) -> [a] -> [[a]]
classify f xs =
    map (map snd)
	$ groupBy (\ (v,_) (w,_) -> v == w)
	$ sortBy (\ (v,_) (w,_) -> compare v w)
	$ map (\x -> (f x, x)) xs

-- | Check if a list is a substring of another list
isSubstringOf :: Eq a => [a] -> [a] -> Bool
isSubstringOf [] _  = True
isSubstringOf _ []  = False
isSubstringOf xs ys = isPrefixOf xs ys || isSubstringOf xs (tail ys)

lookup' :: Eq a => b -> a -> [(a,b)] -> b
lookup' y x m = fromMaybe y $ lookup x m

infixl 4 !?
(!?) :: [a] -> Int -> Maybe a
[] !? _	    = Nothing
(x:xs) !? 0 = Just x
(x:xs) !? n = xs !? (n - 1)

-- | Extract the first element matching the predicate.
extract :: (a -> Bool) -> [a] -> Maybe (a,[a])
extract p [] = Nothing
extract p (x:xs)
    | p x	= Just (x,xs)
    | otherwise	= fmap (\ (y,xs') -> (y,x:xs')) $ extract p xs

